# -*- coding: utf-8 -*-
import argparse
import sys
import pkg_resources
import logging
import inspect
import os
import getpass
import platform
from datetime import datetime
import requests
import time
import pty

from openremote_cli import config
from openremote_cli import scripts
from openremote_cli import gen_aws_smtp_credentials

logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class OpenRemote(object):
    def __init__(self, arguments):
        parser = argparse.ArgumentParser(
            description=f'OpenRemote Command Line Interface (CLI) version {package_version()}',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            #                   conflict_handler='resolve',
        )

        self.base_subparser = self._add_std_arguments(parser)

        # Create subparsers
        self.subparsers = parser.add_subparsers(
            title='command', dest='command'
        )
        for attr in dir(self):
            if inspect.ismethod(getattr(self, attr)) and attr[0] != '_':
                getattr(self, attr)([])

        args, unknown = self.base_subparser.parse_known_args(arguments)
        if args.quiet:
            config.QUIET = True
            args.verbosity = 0

        config.LEVEL = {
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
            3: logging.DEBUG,
        }.get(args.verbosity, logging.DEBUG)
        logging.getLogger().setLevel(config.LEVEL)

        if args.no_telemetry:
            config.TELEMETRY = False

        if args.dry_run is True:
            logging.info('Enabling dry run mode')
            config.DRY_RUN = True
        if args.verbosity > 0:
            config.VERBOSE = True

        logging.debug(args)

        # handle no arguments
        if len(arguments) == 0:
            self.help(['--help'])
        else:
            command = args.command
            logging.debug(f"command: {command}")
            # handle undefined command
            if not command:
                self.help(arguments)
            elif not hasattr(self, command):
                if command[0] != '-':
                    print('Unknown command ' + command)
                self.help(arguments)
            else:
                if not config.QUIET:
                    if config.DRY_RUN:
                        print("--dry-run active!")
                    if not config.VERBOSE:
                        print(
                            "To see commands use -v switch (-vvv for debug)\n"
                        )
                    else:
                        print(
                            'If you need help go to https://forum.openremote.io/\n'
                        )
                    # use dispatch pattern to invoke method with same name so it's
                    # easy to add new subcommands
                    logging.debug('dispatching ' + command + f'({arguments})')
                getattr(self, command)(arguments)

    # Basic command run without arguments adds parser
    def help(self, arguments=[]):
        if len(arguments) > 0:
            self.base_subparser.parse_args(['-h'])
        else:
            pass
            # parser = self.__parser('help', 'CLI help')

    def configure_aws(self, arguments=[]):
        if len(arguments) > 0:
            args = self.base_subparser.parse_args(arguments)
            logging.debug(args)
            if args.id and args.secret:
                scripts.configure_aws(args.id, args.secret, args.region)
            elif args.secret:
                # Calculate SMTP password from AWS secret key
                print(
                    gen_aws_smtp_credentials.calculate_key(
                        args.secret, config.REGION
                    ),
                    end='',
                )
            else:
                scripts.configure_aws_show()
        else:
            parser = self.__parser(
                'configure_aws', 'configure AWS credentials'
            )
            arguments = parser.add_argument_group("configure_aws arguments")
            arguments.add_argument(
                "-i",
                "--id",
                type=str,
                required=False,
                help="Access key ID (if not given then return SMTP password)",
            )
            arguments.add_argument(
                "-s", "--secret", type=str, help="Secret access key"
            )
            arguments.add_argument(
                "-r",
                "--region",
                type=str,
                required=False,
                help="AWS Region",
                default="eu-west-1",
            )

    def map(self, arguments=[]):
        if len(arguments) > 0:
            args = self.base_subparser.parse_args(arguments)
            logging.debug(args)
            if args.action == 'list':
                scripts.map_list()
            elif args.action == 'upload':
                scripts.map_upload(args.file)
            elif args.action == 'download':
                scripts.map_download(args.file)
            elif args.action == 'delete':
                scripts.map_delete(args.file)
            else:
                raise ValueError(f"'{args.action}' not implemented")
        else:
            parser = self.__parser('map', 'map storage/retrieve service')
            arguments = parser.add_argument_group("map arguments")
            arguments.add_argument(
                '-a',
                '--action',
                nargs="?",
                choices=['list', 'upload', 'download', 'delete'],
                help='list/upload/download/delete map from S3',
                default='list',
            )
            arguments.add_argument('-f', '--file', type=str, help="file name")

    def deploy(self, arguments=[]):
        if len(arguments) > 0:
            args = self.base_subparser.parse_args(arguments)
            logging.debug(args)
            if args.action == 'create':
                print(
                    'Deploying OR... This usually takes less than 15 minutes.\n'
                )
                if args.provider == 'aws':
                    scripts.deploy_aws(args.password, args.dnsname)
                else:
                    smtp_user, smtp_password = None, None
                    if args.with_email:
                        logging.debug('Creating SMTP credentials')
                        smtp_user, smtp_password = scripts.smtp_credentials(
                            args.dnsname
                        )
                        logging.debug(
                            f'user: {smtp_user}, password: {smtp_password}'
                        )
                    if args.provider == 'rich':
                        if args.dnsname == 'localhost':
                            raise Exception('--dnsname must set')
                        scripts.deploy_rich(
                            args.password,
                            smtp_user,
                            smtp_password,
                            args.dnsname,
                        )
                    else:
                        scripts.deploy(
                            args.password,
                            smtp_user,
                            smtp_password,
                            args.dnsname,
                        )
            elif args.action == 'remove':
                print('Removing OR stack...\n')
                if args.provider == 'aws':
                    scripts.remove_aws(args.dnsname)
                elif args.provider == 'rich':
                    # TODO
                    raise Exception('Not implemented')
                else:
                    scripts.remove(args.dnsname)
            elif args.action == 'clean':
                print('Cleaning OR resources...\n')
                if args.provider == 'aws':
                    scripts.remove_aws(args.dnsname)
                elif args.provider == 'rich':
                    # TODO
                    raise Exception('Not implemented')
                else:
                    scripts.clean()
            elif args.action == 'health':
                scripts.deploy_health(args.dnsname, args.verbosity)
            else:
                raise ValueError(f"'{args.action}' not implemented")
        else:
            logging.debug('adding deploy parser')
            parser = self.__parser(
                'deploy',
                'Deploy OpenRemote stack. By default create on localhost.',
            )
            arguments = parser.add_argument_group("deploy arguments")
            arguments.add_argument(
                '-a',
                '--action',
                nargs="?",
                choices=['create', 'remove', 'clean', 'health'],
                help='create/remove/clean OpenRemote stack',
                default='create',
            )
            arguments.add_argument(
                '-p',
                '--password',
                type=str,
                default='secret',
                help='password for admin user',
            )
            arguments.add_argument(
                '--provider',
                nargs="?",
                choices=['aws', 'localhost', 'rich'],
                default='localhost',
                help='where the stack should be deployed (rich is on localhost but with artifacts from S3)',
            )
            arguments.add_argument(
                '--dnsname',
                type=str,
                help='host and domain name',
                default='localhost',
            )
            arguments.add_argument(
                '--with-email',
                action='store_true',
                help='generate valid SMTP server access keys',
            )

    def manager(self, arguments=[]):
        if len(arguments) > 0:
            args = self.base_subparser.parse_args(arguments)
            logging.debug(args)
            if not args.password:
                try:
                    args.password = os.getenv('SETUP_ADMIN_PASSWORD')
                except:
                    raise Exception("No password given")
            if args.login:
                scripts.manager_login(args.dnsname, args.user, args.password)
            if args.list_realms is True:
                print('\nListing realms\n--------------')
                scripts.manager_list_realms(args.dnsname)
            if args.list_users is True:
                print(
                    f'\nListing users for {args.realm} realm\n--------------'
                )
                scripts.manager_list_users(args.realm, args.dnsname)
            if args.list_public_assets:
                scripts.manager_list_public_assets(args.realm, args.dnsname)
            if args.create_user is True:
                raise Exception('not implemented')
            if args.delete_user is True:
                raise Exception('not implemented')
        else:
            parser = self.__parser('manager', 'manage online manager')
            arguments = parser.add_argument_group("manager arguments")
            arguments.add_argument(
                '-l', '--login', action='store_true', help='login into manager'
            )
            arguments.add_argument(
                '-u',
                '--user',
                type=str,
                required=False,
                help='username',
                default='admin',
            )
            arguments.add_argument(
                '-p', '--password', required=False, help='user password'
            )
            arguments.add_argument(
                '--dnsname',
                type=str,
                required=False,
                help='OpenRemote dns',
                default='demo.openremote.io',
            )
            arguments.add_argument(
                '--realm',
                required=False,
                type=str,
                help='realm to work on',
                default='master',
            )
            arguments.add_argument(
                '--list-realms',
                required=False,
                action='store_true',
                help='list defined realms',
            )
            arguments.add_argument(
                '--create-realm',
                required=False,
                action='store_true',
                help='create new realm',
            )
            arguments.add_argument(
                '--list-users',
                required=False,
                action='store_true',
                help='list defined users in a realm',
            )
            arguments.add_argument(
                '--create-user',
                required=False,
                action='store_true',
                help='create users in a realm',
            )
            arguments.add_argument(
                '--delete-user',
                required=False,
                action='store_true',
                help='delete users from a realm',
            )
            arguments.add_argument(
                '--list-public-assets',
                required=False,
                action='store_true',
                help='list public assets in a realm',
            )

    def shell(self, arguments=[]):
        if len(arguments) > 0:
            pty.spawn("/bin/sh")
        else:
            self.__parser(
                'shell', 'spawn shell (useful inside of docker container)'
            )

    def perquisites(self, arguments=[]):
        if len(arguments) > 0:
            args = self.base_subparser.parse_args(arguments)
            logging.debug(args)
            if args.install is True:
                print('Checking and installing missing tools.\n')
                logging.error("Not implemented")
            else:
                if not config.QUIET:
                    print('Checking for required tools...\n')
                scripts.check_tools()
        else:
            logging.debug('adding perquisites parser')
            parser = self.__parser(
                'perquisites', 'Check if all required tools are installed'
            )
            arguments = parser.add_argument_group("perquisites arguments")
            arguments.add_argument(
                '--install', action='store_true', help='install missing tools'
            )

    def __parser(self, name, description):
        parser = self.subparsers.add_parser(
            name=name,
            description=description,
            help=description,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            # TODO see what is wrong with parents
            # conflict_handler='resolve',
            # parents=[self.base_subparser],
        )
        return self._add_std_arguments(parser)

    def _add_std_arguments(self, parser):
        parser.add_argument(
            '-V',
            '--version',
            action='version',
            version=f'%(prog)s/{package_version()} {sys.version} {platform.system()}/{platform.version()}',
        )
        parser.add_argument(
            '-n',
            '--dry-run',
            action='store_true',
            help='showing effects without actual run and exit',
        )
        parser.add_argument(
            '-v',
            '--verbosity',
            action='count',
            default=0,
            help='increase output verbosity',
        )
        parser.add_argument(
            '-t',
            '--no-telemetry',
            action='store_true',
            help="Don't send usage data to server",
        )
        parser.add_argument(
            '-q', '--quiet', action='store_true', help='suppress info'
        )
        return parser


def send_metric(cli_input, exit_reason, exit_code, duration):
    input_cmd = " ".join(cli_input)
    if 'configure_aws' in input_cmd:
        # never telemetry secrets!
        return
    try:
        user_id = f'{getpass.getuser()}'
    except:
        # We don't have login in github workflow
        user_id = 'No login!'
    payload = {
        "metrics": [
            {
                "userId": user_id,
                "cliVersion": f'{package_version()}',
                "osPlatform": platform.system(),
                "osVersion": platform.version(),
                "pythonVersion": sys.version,
                "command": {
                    "input": input_cmd,
                    "exitReason": exit_reason,
                    "exitCode": exit_code,
                    "duration": duration,
                    "timestamp": datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
                },
            }
        ]
    }

    try:
        requests.post(config.TELEMETRY_URL, json=payload, timeout=2000)
    except requests.exceptions.RequestException as exception:
        logging.debug(str(exception))


def package_version():
    return pkg_resources.get_distribution('openremote_cli').version


def isLatestVersion():
    if config.QUIET:
        return
    # Check PyPI for the latest version number
    contents = requests.get(f'https://pypi.org/pypi/openremote-cli/json')
    data = contents.json()
    latest_version = data['info']['version']
    if latest_version != package_version():
        print(
            f'\nyour version ({package_version()}) < PyPI version ({latest_version}). Consider\npip3 install --upgrade openremote-cli'
        )


def main():
    config.initialize()
    exit_reason = "program finished normally"
    exit_code = 0
    start = time.time()
    try:
        OpenRemote(sys.argv[1:])
    except Exception as error:
        logging.error(error)
        exit_reason = str(error)
        exit_code = -1
        try:
            # If we made this exception then this should work
            exit_code = error[0]
            exit_reason = error[1]
        except:
            # If not then it is already good enough
            pass
    finally:
        end = time.time()
        if (
            config.TELEMETRY
            and '--no-telemetry' not in sys.argv
            and '-t' not in sys.argv
        ):
            logging.debug(f'Sending telemetry to {config.TELEMETRY_URL}')
            send_metric(sys.argv[1:], exit_reason, exit_code, end - start)
        else:
            logging.debug(f'skipping telemetry: {exit_reason}')
        # Check if there is a new version on PyPI
        isLatestVersion()
        return exit_code


# Support invoking the script directly from source
if __name__ == '__main__':
    exit(main())



#!/usr/bin/env python3

import hmac
import hashlib
import base64
import argparse

SMTP_REGIONS = [
    'us-east-2',  # US East (Ohio)
    'us-east-1',  # US East (N. Virginia)
    'us-west-2',  # US West (Oregon)
    'ap-south-1',  # Asia Pacific (Mumbai)
    'ap-northeast-2',  # Asia Pacific (Seoul)
    'ap-southeast-1',  # Asia Pacific (Singapore)
    'ap-southeast-2',  # Asia Pacific (Sydney)
    'ap-northeast-1',  # Asia Pacific (Tokyo)
    'ca-central-1',  # Canada (Central)
    'eu-central-1',  # Europe (Frankfurt)
    'eu-west-1',  # Europe (Ireland)
    'eu-west-2',  # Europe (London)
    'sa-east-1',  # South America (Sao Paulo)
    'us-gov-west-1',  # AWS GovCloud (US)
]

# These values are required to calculate the signature. Do not change them.
DATE = "11111111"
SERVICE = "ses"
MESSAGE = "SendRawEmail"
TERMINAL = "aws4_request"
VERSION = 0x04


def sign(key, msg):
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()


def calculate_key(secret_access_key, region):
    if region not in SMTP_REGIONS:
        raise ValueError(f"The {region} Region doesn't have an SMTP endpoint.")

    signature = sign(("AWS4" + secret_access_key).encode('utf-8'), DATE)
    signature = sign(signature, region)
    signature = sign(signature, SERVICE)
    signature = sign(signature, TERMINAL)
    signature = sign(signature, MESSAGE)
    signature_and_version = bytes([VERSION]) + signature
    smtp_password = base64.b64encode(signature_and_version)
    return smtp_password.decode('utf-8')


def main():
    parser = argparse.ArgumentParser(
        description='Convert a Secret Access Key for an IAM user to an SMTP password.'
    )
    parser.add_argument('secret', help='The Secret Access Key to convert.')
    parser.add_argument(
        'region',
        help='The AWS Region where the SMTP password will be used.',
        choices=SMTP_REGIONS,
    )
    args = parser.parse_args()
    print(calculate_key(args.secret, args.region))


if __name__ == '__main__':
    main()



#!/usr/bin/python3

import smtplib
import sys
import logging
import argparse

from openremote_cli import config


def sendmail(receiver, subject, message_text, smtp_user, smtp_password):

    message = '\r\n'.join(
        [
            f'From: no-reply@openremote.io',
            f'To: {receiver}',
            f'Subject: {subject}',
            message_text,
        ]
    )

    try:
        server = smtplib.SMTP(config.SMTP_SERVER, 587)
        server.ehlo()
        server.starttls()
        server.login(smtp_user, smtp_password)

        server.sendmail('no-reply@openremote.io', [receiver], message)
        logging.debug('Successfully sent email')
    except Exception as error:
        logging.error(error)
        logging.debug('Error: unable to send email')


def main():
    parser = argparse.ArgumentParser(
        description='Sends e-mail with given SMTP credentials'
    )
    parser.add_argument('user', help='SMTP User ID. e.g. AKI...')
    parser.add_argument('password', help='SMTP password')
    parser.add_argument(
        '-t',
        '--to',
        help='Email receiver',
        required=False,
        default='no-reply@openremote.io',
    )
    parser.add_argument(
        '--subject', help='Email subject', required=False, default='Test mail'
    )
    parser.add_argument('--message', help='Email message body', required=False)

    args = parser.parse_args()
    if not args.message:
        args.message = f"""
    This is a test e-mail message from {args.sender} to {args.to} via AWS.

    User: {args.user}
    Password: {args.password}
    """

    sendmail(
        args.sender,
        args.to,
        args.subject,
        args.message,
        args.user,
        args.password,
    )


if __name__ == '__main__':
    config.initialize()
    main()



# -*- coding: utf-8 -*-
import logging
import subprocess

from openremote_cli import config


def execute(args, no_exception=False, echo=False):
    logging.debug(f'executing command:\n\n\t{args}\n')

    if config.VERBOSE:
        print(f'> {args}')

    if config.DRY_RUN:
        return 0, 'Invoked using dry run'
    else:
        try:
            output = subprocess.run(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,
                # executable='/bin/bash',
            )
        except Exception as exception:
            logging.error(f'Failed to execute {args}')
            if no_exception:
                return exception.errno, exception.strerror
            else:
                raise Exception(exception.errno, exception.strerror)
        if output.returncode != 0:
            if no_exception:
                logging.debug(f'stderr: {output.stdout}')
            else:
                logging.error(f'stderr: {output.stdout}')
                raise Exception(
                    output.returncode, output.stdout.decode('utf-8')
                )
        else:
            logging.debug(f'stdout: {output.stdout}')
        if echo and not config.QUIET:
            print(output.stdout.decode('utf-8'))
        return output.returncode, output.stdout.decode('utf-8')



import os
import configparser
import logging

from pathlib import Path


def _config_file_name():
    return f'{str(Path.home())}/.openremote/config.ini'


def initialize():

    # user persistent config
    global TELEMETRY_URL, REGION, PROFILE, BUCKET, SMTP_SERVER

    config = configparser.ConfigParser()
    config.read(_config_file_name())

    if config.sections() == []:
        try:
            if not os.path.exists(_config_file_name()):
                os.makedirs(
                    f'{str(Path.home())}/.openremote',
                    mode=0o700,
                    exist_ok=True,
                )
                try:
                    config['DEFAULT'] = {
                        'telemetry_url': f"{os.environ['TELEMETRY_URL']}/metrics"
                    }
                except:
                    config['DEFAULT'] = {
                        'telemetry_url': 'https://cli.developers.openremote.io/metrics'
                    }
                config['AWS'] = {
                    'profile': 'openremote-cli',
                    'bucket': 'openremote-mvp-map-storage',
                    'region': 'eu-west-1',
                }
                with open(_config_file_name(), 'w') as conf:
                    config.write(conf)
                    print(f'Config created in {_config_file_name()}')
        except Exception as error:
            logging.error(error)

    default = config['DEFAULT']
    TELEMETRY_URL = default['telemetry_url']
    try:
        if os.environ['TELEMETRY_URL']:
            TELEMETRY_URL = f"{os.environ['TELEMETRY_URL']}/metrics"
    except:
        pass

    # AWS parameters
    aws = config['AWS']
    PROFILE = aws['profile']
    BUCKET = aws['bucket']
    REGION = aws['region']
    SMTP_SERVER = f'email-smtp.{REGION}.amazonaws.com'

    # Runtime config
    global DRY_RUN, LEVEL, VERBOSE, TELEMETRY, QUIET

    VERBOSE = False
    DRY_RUN = False
    TELEMETRY = True
    QUIET = False
    LEVEL = 'logging.ERROR'


def store_token(url, token, refresh):
    config = configparser.ConfigParser()
    config.read(_config_file_name())
    try:
        # first try to update
        managerUrl = config[url]
        managerUrl['access_token'] = token
        managerUrl['refresh_token'] = refresh
    except:
        # if not the create
        config[url] = {'access_token': token, 'refresh_token': refresh}
    with open(_config_file_name(), 'w') as conf:
        config.write(conf)


def get_token(url):
    config = configparser.ConfigParser()
    config.read(_config_file_name())
    # TODO refresh toke or ask to login again
    # managerUrl = config[url]
    return config[url]['access_token']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openremote_cli']

package_data = \
{'': ['*']}

install_requires = \
['emojis>=0.6.0,<0.7.0',
 'python-keycloak>=0.24.0,<0.25.0',
 'requests>=2.25.1,<3.0.0',
 'wget>=3.2,<4.0']

entry_points = \
{'console_scripts': ['openremote-cli = openremote_cli.cli:main',
                     'or = openremote_cli.cli:main']}

setup_kwargs = {
    'name': 'openremote-cli',
    'version': '0.5.8',
    'description': 'OpenRemote Command-line interface',
    'long_description': "![CI/CD](https://github.com/openremote/openremote-cli/workflows/CI/CD/badge.svg)\n![docker](https://github.com/openremote/openremote-cli/workflows/docker/badge.svg)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Open Source? Yes!](https://badgen.net/badge/Open%20Source%20%3F/Yes%21/blue?icon=github)](https://github.com/Naereen/badges/)\n[![Donate](https://img.shields.io/badge/donations-appreciated-green.svg)](https://www.paypal.com/webapps/shoppingcart?flowlogging_id=4b5432e8ad1a8&mfid=1616517920020_4b5432e8ad1a8#/checkout/openButton)\n\n# OpenRemote CLI\n\n## Intro\n\nThis is Command Line Interface to OpenRemote IoT platform. It's purpose is to reduce friction of using OpenRemote by Do It Yourself users.\n\nIt is published as a Python package on [PyPi.org](https://pypi.org/project/openremote-cli/)\n\n## About OR\n\nOpenRemote is a great real OpenSource IoT platform.\n\n## Usage\n\nTo install/upgrade the CLI:\n```bash\npip install --upgrade openremote-cli\nor --version\n```\n\n### Deploy local OpenRemote stack\n\n```bash\nor deploy\n```\n\nWhen it finishes you should be able to open https://localhost and login with admin:secret to the manager. Depending on your machine it can take few minutes to\nstart or services (usually less than 10 minutes). If it does not work yet, check it with ```docker ps``` to see if all services are in healthy or starting stage.\n\n### Remove local OpenRemote stack\n\n```bash\nor deploy --action remove\n```\n\n### Deploy OpenRemote stack on AWS with DNS entry\n\n```bash\nor deploy --platform aws --dnsname myiot.mydomain.com\n```\n### Check health of the openremote stack\n```bash\nor deploy -a health --dnsname myiot.mydomain.com -v\n```\n\n### Install shell completion extension\n\n#### zsh\n\n```bash\nTODO\n```\n\n#### bash\n\n```\nTODO\n```\n\n## Configure existing OpenRemote instance\n\n### Add users\n\n### Add assets\n\n### Add customization\n\n## Develop openremote-cli\n\nFollowing tools are used:\n- python (pip, poetry, black, PyPI.org)\n- git (pre-commit, actions)\n- docker (docker-compose, swarm)\n- AWS (CloudFormation)\n- make\n\n### Adding feature\n\nIn this project we use Behavior-driven development (or BDD). BDD is an agile\nsoftware development technique that encourages collaboration between developers,\nQA and non-technical or business participants in a software project.\n\nThis project uses Gherkin to define what features which should be covered. Features\nfiles can be generated by people on manager level or even higher. An example\nof file defining a feature:\n\n```gherkin\nFeature: deploy\n\n  Scenario: deploy to localhost\n    Given we have docker and docker-compose installed\n    When we call openremote-cli --dry-run deploy --action create\n    Then show what will be done\n```\n\nWhen the feature is implemented it can be checked with behave:\n\n```bash\n> behave\nFeature: deploy # features/deploy.feature:1\n\n  Scenario: deploy to localhost                                  # features/deploy.feature:3\n    Given we have docker and docker-compose installed            # features/steps/deploy_steps.py:8 0.453s\n    When we call openremote-cli --dry-run deploy --action create # features/steps/deploy_steps.py:16 0.591s\n    Then show what will be done                                  # features/steps/deploy_steps.py:24 0.000s\n\n1 feature passed, 0 failed, 0 skipped\n1 scenario passed, 0 failed, 0 skipped\n3 steps passed, 0 failed, 0 skipped, 0 undefined\nTook 0m1.044s\n```\n",
    'author': 'Michal Rutka',
    'author_email': 'michal@openremote.io',
    'maintainer': 'OpenRemote',
    'maintainer_email': 'developers@openremote.io',
    'url': 'https://www.openremote.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import logging
import uuid
import wget
import json
import urllib.request
import os
import string
import requests
import ssl
import time
import emojis
import re


from keycloak import KeycloakOpenID
from random import choice, randint
from openremote_cli import shell
from openremote_cli import config
from openremote_cli import gen_aws_smtp_credentials, email


def deploy(password, smtp_user, smtp_password, dnsname):
    shell.execute('docker swarm init', no_exception=True)
    shell.execute(
        'docker volume rm openremote_postgresql-data', no_exception=True
    )
    # TODO fetch tar from S3 and copy to the docker volume
    # shell.execute('docker volume create openremote_deployment-data')
    # shell.execute(
    #     'docker run --rm -v openremote_deployment-data:/deployment openremote/deployment:mvp'
    # )
    if config.VERBOSE is True:
        print(
            '> wget -nc https://github.com/openremote/openremote/raw/master/mvp/mvp-docker-compose.yml'
        )
    if not config.DRY_RUN:
        wget.download(
            'https://github.com/openremote/openremote/raw/master/mvp/mvp-docker-compose.yml'
        )
    env = ''
    if password != 'secret':
        env = f'PASSWORD={password} '
    if smtp_user and smtp_password:
        env = (
            f'{env}EMAIL_USER={smtp_user} '
            f'EMAIL_PASSWORD={smtp_password} '
            f'EMAIL_HOST=email-smtp.{config.REGION}.amazonaws.com '
        )
    if dnsname != 'localhost':
        identity = dnsname
        if _check_ip(dnsname):  # it is pure IP
            dnsname = 'localhost'  # prevent proxy from issuing cert
        env = f'{env}DOMAINNAME={dnsname} IDENTITY_NETWORK_HOST={identity} '
        # As you may be facing internet change default password for security
        generate_password, password = _password(password)
        if generate_password:
            env = f'{env}PASSWORD={password} '
        # Deploy with docker-compose as proxy container does not obey
        # health check within swarm mode
        # TODO revisit proxy image to fix this, otherwise letsencrypt has problems
        shell.execute(
            f'{env}docker-compose -f mvp-docker-compose.yml -p openremote up -d'
        )
        if not config.DRY_RUN:
            print('\nStack deployed, waiting for startup to complete', end=' ')
            while _deploy_health(dnsname, 0) == 0:
                time.sleep(3)
                print('.', end='', flush=True)
            print(emojis.encode(':thumbsup:\n'))
        if config.VERBOSE is True:
            print(
                f'Open https://{dnsname} and login with admin:{password}\n\n'
                'Remove the stack when you are done:\n'
                '> docker-compose -f mvp-docker-compose.yml -p openremote down\n'
                '> rm mvp-docker-compose.yml\n'
            )
    else:
        shell.execute(
            f'{env}docker stack deploy -c mvp-docker-compose.yml openremote'
        )
        if not config.DRY_RUN:
            print('\nStack deployed, waiting for startup to complete', end=' ')
            c = 0
            # TODO deploying from docker image has problems with health checking hence 10min time out
            while _deploy_health(dnsname, 0) == 0 and c < 200:
                time.sleep(3)
                print('.', end='', flush=True)
                c += 1
            print(emojis.encode(':thumbsup:\n'))
        if config.VERBOSE is True:
            print(
                f'\nOpen https://localhost and login with admin:{password}\n\n'
                'To remove the stack when you are done:\n'
                '> docker stack rm openremote\n'
            )
        if not config.DRY_RUN:
            os.remove(f'mvp-docker-compose.yml')
    if config.VERBOSE is True:
        print(
            'To remove docker resources:\n'
            "> docker images --filter 'reference=openremote/*' -q | xargs docker rmi\n"
            "> docker volume ls --filter 'dangling=true' -q | xargs docker volume rm"
        )


def deploy_health(dnsname, verbosity=0):
    print(_deploy_health(dnsname, verbosity))


def _deploy_health(dnsname, verbosity):
    try:
        # We need this for self-signed certs like localhost
        ssl._create_default_https_context = ssl._create_unverified_context
        health = json.loads(
            urllib.request.urlopen(
                f'https://{dnsname}/api/master/health'
            ).read()
        )
        if verbosity == 0:
            return health['system']['version']
        elif verbosity == 1:
            return health['system']
        else:
            return health
    except:
        if '.' not in dnsname and dnsname != 'localhost':
            host, domain = _split_dns(dnsname)
            deploy_health(f'{host}.{domain}', verbosity)
        elif verbosity == 0:
            return 0
        else:
            return f'Error calling\ncurl https://{dnsname}/api/master/health'


def _split_dns(dnsname):
    if '.' in dnsname:
        host = dnsname.split('.')[0]
        domain = dnsname[len(host) + 1 :]
    else:
        logging.debug('adding default domain')
        host = dnsname
        domain = 'mvp.openremote.io'
    return host, domain


def _password(password):
    generate_password = password == 'secret'
    if generate_password:
        characters = string.ascii_letters + string.digits
        password = "".join(choice(characters) for x in range(randint(8, 16)))
        if not config.QUIET:
            print(f'\nGenerated password: {password}\n')
    return generate_password, password


def deploy_aws(password, dnsname):
    host, domain = _split_dns(dnsname)
    logging.debug(f'{dnsname} => {host} + {domain}')
    stack_name = f'{host}-{uuid.uuid4()}'
    check_aws_perquisites()
    generate_password, password = _password(password)
    if config.VERBOSE is True:
        print(
            '> wget -nc https://github.com/openremote/openremote/raw/master/mvp/aws-cloudformation.template.yml'
        )
    if not config.DRY_RUN:
        wget.download(
            'https://github.com/openremote/openremote/raw/master/mvp/aws-cloudformation.template.yml'
        )
    shell_exec = shell.execute(
        f'aws cloudformation create-stack --stack-name {stack_name} '
        f'--template-body file://aws-cloudformation.template.yml --parameters '
        f'ParameterKey=DomainName,ParameterValue={domain} '
        f'ParameterKey=HostName,ParameterValue={host} '
        f'ParameterKey=HostedZone,ParameterValue=true '
        f'ParameterKey=OpenRemotePassword,ParameterValue={password} '
        f'ParameterKey=InstanceType,ParameterValue=t3a.small '
        f'ParameterKey=KeyName,ParameterValue=openremote '
        f'--capabilities CAPABILITY_NAMED_IAM '
        f'--profile={config.PROFILE}'
    )
    print(f'\n{shell_exec[1]}')
    if shell_exec[0] != 0:
        raise Exception(shell_exec)

    print('Waiting for CloudFormation...')
    # TODO make better feedback to the user
    # code, output = shell.execute
    # CREATE_STACK_STATUS=$(aws --region ${AWS_REGION} --profile ${AWS_PROFILE} cloudformation describe-stacks --stack-name ${STACK_NAME} --query 'Stacks[0].StackStatus' --output text)
    # while [[ $CREATE_STACK_STATUS == "REVIEW_IN_PROGRESS" ]] || [[ $CREATE_STACK_STATUS == "CREATE_IN_PROGRESS" ]]
    # do
    #     # Wait 30 seconds and then check stack status again
    #     sleep 30
    #     CREATE_STACK_STATUS=$(aws --region ${AWS_REGION} --profile ${AWS_PROFILE} cloudformation describe-stacks --stack-name ${STACK_NAME} --query 'Stacks[0].StackStatus' --output text)
    # done
    shell.execute(
        f'aws cloudformation wait stack-create-complete '
        f'--stack-name {stack_name} --profile {config.PROFILE}'
    )
    if not config.DRY_RUN:
        os.remove(f'aws-cloudformation.template.yml')
        if generate_password:
            # In case of password generation get email credentials and send it to support
            code, output = shell.execute(
                f'aws cloudformation describe-stacks --stack-name {stack_name} --profile {config.PROFILE} '
                '--query "Stacks[0].Outputs[?OutputKey==\'UserId\'||OutputKey==\'UserSecret\'].OutputValue" --output json'
            )
            credentials = json.loads(output)
            smtp_user = credentials[0]
            smtp_password = gen_aws_smtp_credentials.calculate_key(
                credentials[1], config.REGION
            )
            email.sendmail(
                "michal.rutka@gmail.com",
                "Openremote password",
                f"""
                A new AWS stack {stack_name} has been just created. Because there was a
                default password used a more secure one was generated. The new admin password is
                {password}
                """,
                smtp_user,
                smtp_password,
            )
            print(
                '\nAn email with generated password was sent to support@openremote.io\n'
            )
    elif generate_password:
        print(
            '\nAn email with generated password would be sent to support@openremote.io\n'
        )
    if not config.DRY_RUN:
        shell.execute(
            f'echo "aws cloudformation delete-stack --stack-name {stack_name} --profile {config.PROFILE}" > aws-delete-stack-{host}.{domain}.sh'
        )
        shell.execute(f'chmod +x aws-delete-stack-{host}.{domain}.sh')
        print('\nStack deployed, waiting for startup to complete', end=' ')
        c = 0
        while _deploy_health(f'{host}.{domain}', 0) == 0 and c < 200:
            time.sleep(3)
            print('.', end='', flush=True)
            c += 1
        print(emojis.encode(':thumbsup:\n'))
    print(
        emojis.encode(
            '\nMind that running it cost money :moneybag::moneybag::moneybag:! To free resources execute:\n\n'
            f'aws cloudformation delete-stack --stack-name {stack_name} --profile {config.PROFILE}\n\n'
            'check running stack with health command:\n'
            f'or deploy -a health --dnsname {host}.{domain} -v'
        )
    )


def remove_aws(dnsname):
    check_aws_perquisites()
    shell.execute(f'sh aws-delete-stack-{dnsname}.sh')
    if config.VERBOSE:
        print(f'rm aws-delete-stack-{dnsname}.sh')
    if not config.DRY_RUN:
        os.remove(f'aws-delete-stack-{dnsname}.sh')


def remove(dnsname):
    if dnsname == 'localhost':
        shell.execute(f'docker stack rm openremote')
    else:
        if config.VERBOSE is True:
            print(
                '> wget -nc https://github.com/openremote/openremote/raw/master/mvp/mvp-docker-compose.yml'
            )
        if not config.DRY_RUN:
            wget.download(
                'https://github.com/openremote/openremote/raw/master/mvp/mvp-docker-compose.yml'
            )
        shell.execute(
            f'docker-compose -f mvp-docker-compose.yml -p openremote down'
        )


def clean():
    shell.execute(
        'docker volume rm --force openremote_deployment-data openremote_postgresql-data openremote_proxy-data'
    )
    shell.execute(
        'docker rmi openremote/manager-swarm openremote/deployment '
        'openremote/keycloak openremote/postgresql openremote/proxy '
        'openremote/manager',
        no_exception=True,
    )
    shell.execute('docker system prune --force')


def configure_aws(id, secret, region):
    config.REGION = region
    shell.execute(
        f'aws configure set profile.{config.PROFILE}.aws_access_key_id {id}',
        echo=config.VERBOSE,
    )
    shell.execute(
        f'aws configure set profile.{config.PROFILE}.aws_secret_access_key {secret}',
        echo=config.VERBOSE,
    )
    shell.execute(
        f'aws configure set profile.{config.PROFILE}.region {region}',
        echo=config.VERBOSE,
    )


def configure_aws_show():
    if not config.QUIET:
        print(f'S3 bucket: s3://{config.BUCKET}\n')
        region = shell.execute(
            f'aws configure get region --profile {config.PROFILE}'
        )[1][:-1]
        id = shell.execute(
            f'aws configure get aws_access_key_id --profile {config.PROFILE}'
        )[1][:-1]
        secret = shell.execute(
            f'aws configure get aws_secret_access_key --profile {config.PROFILE}'
        )[1][:-1]
        print(f'--id={id} --secret="{secret}" --region={region}')


def map_upload(path):
    shell.execute(
        f'aws s3 cp {path} s3://{config.BUCKET}/{path} --profile {config.PROFILE}',
        echo=True,
    )
    shell.execute(
        'aws s3api put-object-tagging --bucket %s --key %s --tagging \'TagSet=[{Key=type,Value=deployment-data}]\' --profile %s'
        % (config.BUCKET, path, config.PROFILE),
        echo=True,
    )


def map_list():
    shell.execute(
        f'aws s3 ls s3://{config.BUCKET} --recursive --human-readable  --summarize --profile {config.PROFILE}',
        echo=True,
    )


def map_download(path):
    shell.execute(
        f'aws s3 cp s3://{config.BUCKET}/{path} {path} --profile {config.PROFILE}',
        echo=True,
    )


def map_delete(path):
    shell.execute(
        f'aws s3 rm s3://{config.BUCKET}/{path} --profile {config.PROFILE}',
        echo=True,
    )


def check_tools():
    code, output = shell.execute('docker --version')
    if not config.QUIET:
        print(output)
    code, output = shell.execute('docker-compose --version')
    if not config.QUIET:
        print(output)
    # print('Checking AWS perquisites')
    code, output = shell.execute('aws --version')
    if not config.QUIET:
        print(output)
    return check_aws_perquisites()


def check_aws_perquisites():
    # Check AWS profile
    code, output = shell.execute('aws configure list-profiles')
    if not config.DRY_RUN and config.PROFILE not in output:
        msg = f"aws-cli profile '{config.PROFILE}' missing"
        raise Exception(-1, msg)
    # Check EC2 key
    code, output = shell.execute(
        f'aws ec2 describe-key-pairs --key-names openremote --profile {config.PROFILE}'
    )
    if not config.DRY_RUN and 'openremote' not in output:
        msg = f"ERROR: Missing EC2 keypair 'openremote' in {config.REGION} region (Ireland)"
        raise Exception(-1, msg)
    return True


# E-mail credentials
def smtp_credentials(dnsname):
    if dnsname == 'localhost':
        user_name = f'ses-user-{dnsname}-{uuid.uuid4()}'
    else:
        user_name = f'ses-user-{dnsname}'
    code, output = shell.execute(
        f'aws iam create-user --user-name {user_name} --profile {config.PROFILE} --output json'
    )
    logging.debug(output)
    code, output = shell.execute(
        'aws iam put-user-policy --policy-document \'{"Version": "2012-10-17",'
        '"Statement":[{"Effect": "Allow","Action": "ses:SendRawEmail","Resource": "*"}]}\''
        f' --policy-name OpenRemoteSendEmail --user-name {user_name} --profile {config.PROFILE}'
    )
    code, output = shell.execute(
        f'aws iam create-access-key --user-name {user_name} --profile {config.PROFILE} --output json'
    )
    if config.DRY_RUN:
        return 'AccessKeyId', 'SecretAccessKey'
    else:
        logging.debug(output)
        access = json.loads(output)
        return (
            access['AccessKey']['AccessKeyId'],
            gen_aws_smtp_credentials.calculate_key(
                access['AccessKey']['SecretAccessKey'], config.REGION
            ),
        )


# Manager
def manager_login(url, username, password):
    keycloak_openid = KeycloakOpenID(
        server_url=f'https://{url}/auth/',
        client_id="admin-cli",
        realm_name="master",
        client_secret_key="secret",
    )
    response = keycloak_openid.token(username, password)
    config.store_token(
        url, response['access_token'], response['refresh_token']
    )


def manager_list_realms(dnsname):
    response = requests.get(
        f'https://{dnsname}/auth/admin/realms',
        headers={'Authorization': f'Bearer ' + config.get_token(dnsname)},
    )
    for r in response.json():
        if config.QUIET:
            print(f"{r['realm']}  \t{r['displayName']}")
        else:
            print(json.dumps(r, indent=2))


def manager_list_users(realm, dnsname):
    response = requests.get(
        f'https://{dnsname}/auth/admin/realms/{realm}/users',
        headers=_bearer(dnsname),
    )
    for r in response.json():
        if config.QUIET:
            print(f"{r['firstName']} {r['lastName']}\t-\t{r['username']}")
        else:
            print(json.dumps(r, indent=2))


def manager_list_public_assets(realm, dnsname):
    response = requests.post(
        f'https://{dnsname}/api/{realm}/asset/public/query',
        # headers=_bearer(dnsname), # TODO why it gives 401 with smartcity realm and not with master?
        json={
            "select": {"include": "ALL_EXCEPT_PATH"},
            "type": {
                "predicateType": "string",
                "value": "urn:openremote:asset:thing",
            },
        },
    )
    for r in response.json():
        if config.QUIET:
            print(f"{r['type']}\t{r['name']}")
        else:
            print(json.dumps(r, indent=2))


def _bearer(dnsname):
    return {'Authorization': f'Bearer ' + config.get_token(dnsname)}


def _check_ip(ip):
    regex = "^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$"
    return re.search(regex, ip)


# Richard's way
def deploy_rich(password, smtp_user, smtp_password, project):
    staging = False
    if project.find('staging.') == 0:
        staging = True
        project = project[8:]
    shell.execute(
        f'aws s3 cp s3://{config.BUCKET}/{project} {project} --recursive --profile {config.PROFILE}'
    )
    shell.execute(f'tar xvf {project}/deployment.tar.gz')
    shell.execute(f'mv mapdata.mbtiles deployment/map/', no_exception=True)
    env = ''
    # TODO reuse installed credentials docker run {project_manager_1} env
    if password != 'secret':
        env = f'PASSWORD={password} '
    if smtp_user and smtp_password:
        env = (
            f'{env}EMAIL_USER={smtp_user} '
            f'EMAIL_PASSWORD={smtp_password} '
            f'EMAIL_HOST=email-smtp.{config.REGION}.amazonaws.com '
        )
    dnsname = f'{project}.openremote.io'
    if staging:
        env = f'{env}DEPLOYMENT_NAME=staging.{project} '
    else:
        env = f'{env}DEPLOYMENT_NAME={project} '
    generate_password, password = _password(password)
    if generate_password:
        env = f'{env}PASSWORD={password} '
    shell.execute(f'{env}docker-compose up -d')
    if not config.DRY_RUN:
        print('\nStack deployed, waiting for startup to complete', end=' ')
        while _deploy_health(dnsname, 0) == 0:
            time.sleep(3)
            print('.', end='', flush=True)
        print(emojis.encode(':thumbsup:'))
    if config.VERBOSE is True:
        print(f'\nOpen https://{dnsname} and login with admin:{password}')
